#include "elevator.h"
#include <QDebug>
#include "eventlogger.h"

// Elevator class handles elevator ui and movement
Elevator::Elevator(int id, int capacity, QPushButton* openButton, QPushButton* closeButton, QPushButton* helpButton, QPushButton* obsButton, QLCDNumber* capacityDisplay, QLabel* elevatorDisplay, QObject *parent)
    : QObject(parent), id(id), currentFloor(0), doorOpen(false), load(0), capacity(capacity), openButton(openButton), closeButton(closeButton), helpButton(helpButton), obsButton(obsButton), capacityDisplay(capacityDisplay), elevatorDisplay(elevatorDisplay), floorButtonStates(7, false), elevatorState(ElevatorState::LOADING), doorState(DoorState::CLOSED)
{
    // Initialize any other necessary handlers
    connect(openButton, &QPushButton::clicked, this, &Elevator::openDoor);
    connect(closeButton, &QPushButton::clicked, this, &Elevator::closeDoor);
    connect(obsButton, &QPushButton::toggled, this, &Elevator::setDoorObstacle);
    connect(helpButton, &QPushButton::clicked, this, [this]() { handleAlarm(AlarmType::HELP); });

    connect(&doorOpenTimer, &QTimer::timeout, this, [=]() { closeDoor(); });
    connect(&moveTimer, &QTimer::timeout, this, &Elevator::moveElevator);
    connect(&helpEmergTimer, &QTimer::timeout, this, [=]() {
        elevatorDisplay->setText(helpText);
        helpText = "HELP: Connected to 911 Emergency Services";
    });

    elevatorDirection = ElevatorDirection::UP;

    helpText = "HELP: Connected to 911 Emergency Services";
    helpEmergTimer.setInterval(5000);

    doorOpenTimer.setInterval(10000);

    updateDisplay();
}

Elevator::~Elevator()
{
}

int Elevator::getId() const
{
    return id;
}

int Elevator::getCurrentFloor() const
{
    return currentFloor;
}

QVector<int>& Elevator::getDestinationFloors()
{
    return destinationFloors;
}

bool Elevator::isDoorOpen() const
{
    return doorOpen;
}

int Elevator::getLoad() const
{
    return load;
}

int Elevator::getCapacity() const
{
    return capacity;
}

void Elevator::setCapacity(int cap)
{
    capacity = cap;
    emit loadChanged(load);
}

bool Elevator::inEmergency() const
{
    return elevatorState == ElevatorState::EMERGFIRE || elevatorState == ElevatorState::EMERGHELP || elevatorState == ElevatorState::EMERGPOWER;
}

QVector<bool> Elevator::getFloorButtonStates() const
{
    return floorButtonStates;
}

ElevatorState Elevator::getElevatorState() const
{
    return elevatorState;
}

DoorState Elevator::getDoorState() const
{
    return doorState;
}

ElevatorDirection Elevator::getElevatorDirection() const
{
    return elevatorDirection;
}

void Elevator::respondToHelp()
{
    if(helpEmergTimer.isActive()) {
        helpEmergTimer.stop();
        helpText = "HELP: Connected to Building Services.";
        helpEmergTimer.setInterval(1000);
        helpEmergTimer.start();
    }
}

QString Elevator::getStateString(ElevatorState state) const
{
    switch (state) {
    case ElevatorState::MOVING:
        return "Moving";
    case ElevatorState::STALLED:
        return "Stalled";
    case ElevatorState::OVERLOAD:
        return "!OVERLOAD!";
    case ElevatorState::LOADING:
        return "Loading";
    default:
        return "Unknown";
    }
}

QString Elevator::getDoorStateString(DoorState state) const
{
    switch (state) {
    case DoorState::OPEN:
        return "Door Open";
    case DoorState::CLOSED:
        return "Door Closed";
    case DoorState::OBSTRUCTED:
        return "!DOOR OBSTRUCTED!";
    default:
        return "Unknown";
    }
}

QString Elevator::getDirectionString(ElevatorDirection state) const
{
    switch (state) {
    case ElevatorDirection::UP:
        return "UP";
    case ElevatorDirection::DOWN:
        return "DOWN";
    default:
        return "NONE";
    }
}

const QVector<Passenger*>& Elevator::getOccupants() const
{
    return occupants;
}

void Elevator::setDirection(ElevatorDirection dir)
{
    elevatorDirection = dir;
}

void Elevator::addOccupant(Passenger* p)
{
    occupants.append(p);
    load = occupants.length();
    capacityDisplay->display(load);
    emit loadChanged(load);
    addDestinationFloor(p->getDestinationFloor());
}

void Elevator::removeOccupant(Passenger* p)
{
    occupants.removeOne(p);
    load = occupants.length();
    capacityDisplay->display(load);
    emit loadChanged(load);
}

void Elevator::moveToFloor(int floor)
{
    if (elevatorState == ElevatorState::OVERLOAD || elevatorState == ElevatorState::LOADING) return;
    if (doorState != DoorState::CLOSED) return;

    setElevatorState(ElevatorState::MOVING);
    addDestinationFloor(floor);
    sortDestinationFloors();

    if (!moveTimer.isActive())
        moveTimer.start(1000);
}

void Elevator::addDestinationFloor(int floor)
{
    if (!destinationFloors.contains(floor)) {
        destinationFloors.append(floor);
        EventLogger::getInstance().logEvent(QString("Elevator %1 added new floor %2 to queue").arg(id + 1).arg(floor + 1));
    }
}

void Elevator::sortDestinationFloors()
{
    std::sort(destinationFloors.begin(), destinationFloors.end(), [this](int a, int b) {
        if (elevatorDirection == ElevatorDirection::UP)
            return a < b;
        else
            return a > b;
    });
}

void Elevator::moveElevator()
{
    if ((elevatorState == ElevatorState::MOVING || elevatorState == ElevatorState::EMERGFIRE || elevatorState == ElevatorState::EMERGPOWER) && !destinationFloors.isEmpty()) {
        int nextFloor = destinationFloors.first();
        if (currentFloor < nextFloor) {
            elevatorDirection = ElevatorDirection::UP;
            currentFloor++;
        } else if (currentFloor > nextFloor) {
            elevatorDirection = ElevatorDirection::DOWN;
            currentFloor--;
        }
        updateDisplay();
        if (currentFloor == nextFloor) {
            emit arrivedAtFloor(currentFloor);
            destinationFloors.removeFirst();
            if (destinationFloors.isEmpty()) {
                moveTimer.stop();
            }
        }
    }
}

void Elevator::startMovement()
{
    if (!destinationFloors.isEmpty()) {
        if (!moveTimer.isActive()) {
            moveTimer.start(1000);
        }
    } else {
        moveTimer.stop();
    }
}

void Elevator::openDoor()
{
    EventLogger::getInstance().logEvent(QString("Elevator %1 open door call").arg(id + 1));
    if(elevatorState == ElevatorState::LOADING || elevatorState == ElevatorState::OVERLOAD) {
        doorOpen = true;
        setDoorState(DoorState::OPEN);
        if (!doorOpenTimer.isActive()) {
            doorOpenTimer.start();
        }
        emit doorStatusChanged(true);
    }
}

void Elevator::closeDoor()
{
    EventLogger::getInstance().logEvent(QString("Elevator %1 close door call").arg(id + 1));
    if(doorOpenTimer.isActive()) {
        doorOpenTimer.stop();
    }
    if(doorState == DoorState::OBSTRUCTED) return;
    if(elevatorState == ElevatorState::LOADING || elevatorState == ElevatorState::OVERLOAD) {
        doorOpen = false;
        setDoorState(DoorState::CLOSED);
        emit doorStatusChanged(false);
    }
}

void Elevator::setDoorObstacle(bool toggle) {
    if(doorState == DoorState::CLOSED) {
        obsButton->setChecked(false);
        return;
    }
    if(toggle) {
        setDoorState(DoorState::OBSTRUCTED);
        EventLogger::getInstance().logEvent(QString("Elevator %1 door Obstruction").arg(id + 1));
    } else {
        EventLogger::getInstance().logEvent(QString("Elevator %1 door cleared Obstruction").arg(id + 1));
        openDoor();
    }
}

void Elevator::handleAlarm(AlarmType alarmType)
{
    if(alarmType == AlarmType::HELP) {
        if (elevatorState == ElevatorState::EMERGHELP) {
            EventLogger::getInstance().logEvent(QString("Elevator %1 HELP was Cancelled").arg(id + 1));
            setElevatorState(ElevatorState::LOADING);
        } else {
            setElevatorState(ElevatorState::EMERGHELP);
            EventLogger::getInstance().logEvent(QString("Elevator %1 HELP was Activated").arg(id + 1));
            if(helpEmergTimer.isActive()) helpEmergTimer.stop();
        }
    }
    emit alarmTriggered(alarmType);
}

void Elevator::updateDisplay()
{
    switch (elevatorState) {
    case ElevatorState::EMERGHELP:
        elevatorDisplay->setText("HELP: Connecting...");
        helpEmergTimer.start();
        return;
    case ElevatorState::EMERGPOWER:
        elevatorDisplay->setText("[" + QString::number(currentFloor + 1) + "] POWER OUTTAGE TO FLOOR " + QString::number(destinationFloors[0] + 1));
        return;
    case ElevatorState::EMERGFIRE:
        elevatorDisplay->setText("[" + QString::number(currentFloor + 1) + "] FIRE TO FLOOR " + QString::number(destinationFloors[0] + 1) + " DISEMBARK ASAP");
        return;
    default:
        elevatorDisplay->setText("[" + QString::number(currentFloor + 1) + "] - " + getStateString(elevatorState) + " - " + getDoorStateString(doorState) + " [" + getDirectionString(elevatorDirection) + "]");
        return;
    }
}

void Elevator::updateAudioSystem()
{
    // If we need audio functionality
}

void Elevator::pressFloorButton(int floor)
{
    if (floor >= 0 && floor < floorButtonStates.size()) {
        floorButtonStates[floor] = true;
        // Add the floor to the destination floors if it's not already there
        if (!destinationFloors.contains(floor))
            destinationFloors.append(floor);
    }
}

void Elevator::releaseFloorButton(int floor)
{
    if (floor >= 0 && floor < floorButtonStates.size()) {
        floorButtonStates[floor] = false;
    }
}

void Elevator::setElevatorState(ElevatorState state)
{
    if(state == ElevatorState::MOVING && doorState != DoorState::CLOSED) return;
    elevatorState = state;
    emit elevatorStateChanged(state);
    updateDisplay();

    if (state == ElevatorState::MOVING || state == ElevatorState::EMERGPOWER || state == ElevatorState::EMERGFIRE) {
        startMovement();
    }
}

void Elevator::setDoorState(DoorState state)
{
    doorState = state;
    emit doorStateChanged(state);
    updateDisplay();
}
