#ifndef ELEVATOR_H
#define ELEVATOR_H

#include <QObject>
#include <QVector>
#include <QPushButton>
#include <QLCDNumber>
#include <QLabel>
#include <QTimer>
#include "alarmtype.h"
#include "passenger.h"

enum class ElevatorState { MOVING, STALLED, OVERLOAD, LOADING, EMERGFIRE, EMERGHELP, EMERGPOWER };
enum class DoorState { OPEN, CLOSED, OBSTRUCTED };
enum class ElevatorDirection { UP, DOWN };

class Elevator : public QObject
{
    Q_OBJECT

public:
    Elevator(int id, int capacity, QPushButton* openButton, QPushButton* closeButton, QPushButton* helpButton, QPushButton* obsButton, QLCDNumber* capacityDisplay, QLabel* elevatorDisplay, QObject *parent = nullptr);
    ~Elevator();

    int getId() const;
    int getCurrentFloor() const;
    QVector<int>& getDestinationFloors();
    bool isDoorOpen() const;
    int getLoad() const;
    int getCapacity() const;
    void setCapacity(int cap);
    QVector<bool> getFloorButtonStates() const;
    ElevatorState getElevatorState() const;
    ElevatorDirection getElevatorDirection() const;
    DoorState getDoorState() const;
    QString getStateString(ElevatorState state) const;
    QString getDoorStateString(DoorState state) const;
    QString getDirectionString(ElevatorDirection state) const;
    void addOccupant(Passenger* p);
    void removeOccupant(Passenger* p);
    const QVector<Passenger*>& getOccupants() const;
    bool inEmergency() const;
    void respondToHelp();
    void addDestinationFloor(int floor);
    void sortDestinationFloors();
    void setDirection(ElevatorDirection dir);

public slots:
    void moveToFloor(int floor);
    void openDoor();
    void closeDoor();
    void handleAlarm(AlarmType alarmType);
    void updateDisplay();
    void updateAudioSystem();
    void pressFloorButton(int floor);
    void releaseFloorButton(int floor);
    void setElevatorState(ElevatorState state);
    void setDoorState(DoorState state);
    void setDoorObstacle(bool toggle);
private slots:
    void moveElevator();

signals:
    void arrivedAtFloor(int floor);
    void doorStatusChanged(bool open);
    void loadChanged(int load);
    void alarmTriggered(AlarmType alarmType);
    void elevatorStateChanged(ElevatorState state);
    void doorStateChanged(DoorState state);

private:
    int id;
    int currentFloor;
    QVector<int> destinationFloors;
    bool doorOpen;
    int load;
    int capacity;
    QPushButton* openButton;
    QPushButton* closeButton;
    QPushButton* helpButton;
    QPushButton* obsButton;
    QLCDNumber* capacityDisplay;
    QLabel* elevatorDisplay;
    QVector<bool> floorButtonStates;
    ElevatorState elevatorState;
    DoorState doorState;
    ElevatorDirection elevatorDirection;
    QVector<Passenger*> occupants;

    QTimer helpEmergTimer;
    QString helpText;
    QTimer doorOpenTimer;
    QTimer moveTimer;

    void startMovement();
};

#endif // ELEVATOR_H
