#include "elevatorcontrolsystem.h"
#include "eventlogger.h"
#include <QRandomGenerator>

// CENTRALIZE ECS IMPLEMENTATION
ElevatorControlSystem::ElevatorControlSystem(QVector<Elevator*> elevators, QVector<Floor*> floors, QVector<Passenger*> passengers, QObject *parent)
    : QObject(parent), elevators(elevators), floors(floors), passengers(passengers)
{
    // Connect signals from Elevator, Floor, and Passenger objects to the corresponding slots
    // This is basically just observer patterns to make my life easy
    // It meant less code but a headache to debug
    for (Elevator* elevator : elevators) {
        connect(elevator, &Elevator::arrivedAtFloor, this, [=](int floor) { handleElevatorArrivedAtFloor(elevator, floor); });
        connect(elevator, &Elevator::doorStatusChanged, this, [=](bool open) { handleElevatorDoorStatusChanged(elevator, open); });
        connect(elevator, &Elevator::loadChanged, this,  [=](int load) { handleElevatorLoadChanged(elevator, load); });
        connect(elevator, &Elevator::alarmTriggered, this, [=](AlarmType alarmType) { handleElevatorAlarmTriggered(elevator, alarmType); });
    }

    for (Floor* floor : floors) {
        connect(floor, &Floor::upButtonPressed, this, [=](int floorNumber) { handleFloorUpButtonPressed(floor, floorNumber); });
        connect(floor, &Floor::downButtonPressed, this, [=](int floorNumber) { handleFloorDownButtonPressed(floor, floorNumber); });
    }

    for (Passenger* passenger : passengers) {
        connect(passenger, &Passenger::destinationFloorChanged, this, [=](int floor) { handlePassengerDestinationFloorChanged(passenger, floor); });
        connect(passenger, &Passenger::boardingElevator, this, [=](int elevator) { handlePassengerBoardingElevator(passenger, elevator); });
        connect(passenger, &Passenger::exitingElevator, this, [=](int elevator) { handlePassengerExitingElevator(passenger, elevator); });
    }

    for (Floor* floor : floors) {
        for (Passenger* passenger : passengers) {
            if(floor->getFloorNumber() == passenger->getCurrentFloor()) {
                floor->addPerson(passenger);
            }
        }
    }
}

ElevatorControlSystem::~ElevatorControlSystem()
{
}

const QVector<Elevator*>& ElevatorControlSystem::getElevators() const
{
    return elevators;
}

void ElevatorControlSystem::handleElevatorArrivedAtFloor(Elevator* elevator, int floor)
{
    if (elevator) {
        auto elevatorPeople = elevator->getOccupants();
        for (auto it = elevatorPeople.begin(); it != elevatorPeople.end(); ++it) {
            Passenger* p = *it;
            p->setCurrentFloor(floor);
            if (p->getDestinationFloor() == elevator->getCurrentFloor()) {
                elevator->setElevatorState(ElevatorState::LOADING);
            }
        }

        // Check if this floor is a destination floor
        if (elevator->getDestinationFloors().contains(floor)) {
            elevator->setElevatorState(ElevatorState::LOADING);
            elevator->openDoor();
        }

        // Edge cases with limit can cause boarding issues
        if (floor == floors.length() - 1) {
            elevator->setDirection(ElevatorDirection::DOWN);
        } else if (floor == 0 ) {
            elevator->setDirection(ElevatorDirection::UP);
        }
    }
}
void ElevatorControlSystem::handleElevatorDoorStatusChanged(Elevator* elevator, bool open)
{
    if (open && elevator) {
        // Passengers exiting the elevator
        auto elevatorPeople = elevator->getOccupants();
        for (auto it = elevatorPeople.begin(); it != elevatorPeople.end(); ++it) {
            Passenger* p = *it;
            if (p->getDestinationFloor() == elevator->getCurrentFloor()) {
                p->exitElevator(elevator->getId());
            }
        }

        // Passengers boarding the elevator
        Floor* stoppedAtFloor = floors[elevator->getCurrentFloor()];
        auto floorPeople = stoppedAtFloor->getPeople();
        for (auto it = floorPeople.begin(); it != floorPeople.end(); ++it) {
            Passenger* p = *it;
            if (shouldBoardElevator(p, elevator)) {
                p->boardElevator(elevator->getId());
            }
        }

        // Reset floor buttons on arrival
        if (elevator->getElevatorDirection() == ElevatorDirection::UP) {
            stoppedAtFloor->resetUpButtons();
        } else if (elevator->getElevatorDirection() == ElevatorDirection::DOWN) {
            stoppedAtFloor->resetDownButtons();
        }
    } else if (!open && elevator) {
        // Door closed okay we are good to move
        elevator->sortDestinationFloors();
        if (!elevator->getDestinationFloors().isEmpty()) {
            elevator->setElevatorState(ElevatorState::MOVING);
        }
    }
}

// Should passenger board a given elevator?
bool ElevatorControlSystem::shouldBoardElevator(Passenger* passenger, Elevator* elevator) const
{
    const int currentFloor = passenger->getCurrentFloor();
    const int destinationFloor = passenger->getDestinationFloor();

    // If no floors to go to direction doesn't matter!
    if(elevator->getDestinationFloors().isEmpty() && currentFloor != destinationFloor) return true;

    if (elevator->getElevatorDirection() == ElevatorDirection::DOWN) {
        return currentFloor > destinationFloor;
    } else {
        return currentFloor < destinationFloor;
    }
}

// Over capacity check handler
void ElevatorControlSystem::handleElevatorLoadChanged(Elevator* elevator, int load)
{
    if(elevator->getCapacity() < load) {
        elevator->setElevatorState(ElevatorState::OVERLOAD);
        EventLogger::getInstance().logEvent(QString("Elevator %1 is OVERLOADED by %2").arg(elevator->getId() + 1).arg(load - elevator->getCapacity()));
    } else {
        elevator->setElevatorState(ElevatorState::LOADING);
    }
}

// check for raised alarms
void ElevatorControlSystem::handleElevatorAlarmTriggered(Elevator* elevator, AlarmType alarmType)
{
    if(alarmType == AlarmType::HELP) {
        if(QRandomGenerator::global()->bounded(2) == 0) {
            //Building responds to help
            elevator->respondToHelp();
        }
    }
    handleAlarm(alarmType);
}

// Handle other alarms (mostly so the mainwindow can call this function)
void ElevatorControlSystem::handleAlarm(AlarmType alarmType)
{
    int safeFloor = 0;

    switch (alarmType) {
    case AlarmType::POWEROUTAGE:
        // Let off at current floor
        safeFloor = QRandomGenerator::global()->bounded(0, floors.length());
        for (auto it = elevators.begin(); it != elevators.end(); ++it) {
            Elevator* e = *it;
            safeFloor = e->getCurrentFloor();
            e->getDestinationFloors().clear();
            e->addDestinationFloor(safeFloor);
            e->setDoorState(DoorState::CLOSED);
            for(auto jt = e->getOccupants().begin(); jt != e->getOccupants().end(); ++jt) {
                Passenger* p = *jt;
                p->setDestinationFloor(safeFloor);
                p->lock();
            }
            e->setElevatorState(ElevatorState::EMERGPOWER);
        }
        return;
    case AlarmType::FIRE:
        // Pick a random "safe" floor and let everybody off
        safeFloor = QRandomGenerator::global()->bounded(0, floors.length());
        for (auto it = passengers.begin(); it != passengers.end(); ++it) {
            Passenger* p = *it;
            p->setDestinationFloor(safeFloor);
            p->lock();
        }
        for (auto it = elevators.begin(); it != elevators.end(); ++it) {
            Elevator* e = *it;
            e->getDestinationFloors().clear();
            e->addDestinationFloor(safeFloor);
            e->setDoorState(DoorState::CLOSED);
            e->setElevatorState(ElevatorState::EMERGFIRE);
        }
        return;
    default:
        return;
    }
}

// Resets some state
void ElevatorControlSystem::reset() {
    for (auto it = passengers.begin(); it != passengers.end(); ++it) {
        Passenger* p = *it;
        p->setDestinationFloor(0);
        p->setCurrentFloor(0);
        p->updateButtonsAndIndicators(1);
    }
    for (auto it = elevators.begin(); it != elevators.end(); ++it) {
        Elevator* e = *it;
        e->setElevatorState(ElevatorState::LOADING);
        e->closeDoor();
        e->moveToFloor(1);
    }
}

// Dispatch elevator to an UP call
void ElevatorControlSystem::handleFloorUpButtonPressed(Floor* floor, int floorNumber)
{
    dispatchElevator(floorNumber, ElevatorDirection::UP);
}

// Dispatch elevator to a DOWN call
void ElevatorControlSystem::handleFloorDownButtonPressed(Floor* floor, int floorNumber)
{
    dispatchElevator(floorNumber, ElevatorDirection::DOWN);
}

// Best elevator algortithm
void ElevatorControlSystem::dispatchElevator(int floorNumber, ElevatorDirection direction)
{
    Elevator* bestElevator = nullptr;
    int bestDistance = std::numeric_limits<int>::max();

    for (Elevator* elevator : elevators) {
        // Avoid bad states
        if (elevator->getElevatorState() != ElevatorState::OVERLOAD &&
            elevator->getElevatorState() != ElevatorState::EMERGFIRE &&
            elevator->getElevatorState() != ElevatorState::EMERGPOWER &&
            elevator->getElevatorState() != ElevatorState::EMERGHELP &&
            elevator->getDoorState() != DoorState::OBSTRUCTED) {

            int distance;
            // Find smallest distance based on direction and current location away from floor
            if (elevator->getElevatorDirection() == direction) {
                distance = std::abs(elevator->getCurrentFloor() - floorNumber);
            } else {
                distance = std::abs(elevator->getCurrentFloor() - floorNumber) + floors.size();
            }

            if (distance < bestDistance) {
                bestDistance = distance;
                bestElevator = elevator;
            }
        }
    }

    // Dispatch best elevator
    if (bestElevator) {
        bestElevator->addDestinationFloor(floorNumber);
        bestElevator->sortDestinationFloors();
        bestElevator->setElevatorState(ElevatorState::MOVING);
    }
}

void ElevatorControlSystem::handlePassengerDestinationFloorChanged(Passenger* passenger, int floor)
{
    // On elevator (simulate like just pressed an elevator button add it to destination)
    if (passenger && passenger->getBoardedElevator() >= 0) {
        Elevator* elevator = elevators[passenger->getBoardedElevator()];
        if (elevator) {
            // Are we already on the floor with an open door as the newly selected floor
            if (elevator->getCurrentFloor() == passenger->getDestinationFloor() && elevator->getDoorState() == DoorState::OPEN) {
                passenger->exitElevator(elevator->getId());
            } else {
                elevator->addDestinationFloor(passenger->getDestinationFloor());
                elevator->sortDestinationFloors();
            }
        }
    // if on floor and newly selects check if any elevator can accomodate
    } else if (passenger) {
        for (auto it = elevators.begin(); it != elevators.end(); ++it) {
            Elevator* e = *it;
            if (e->getDoorState() == DoorState::OPEN && shouldBoardElevator(passenger, e)) {
                passenger->boardElevator(e->getId());
                return;
            }
        }
    }
}

// Move passenger over as elevator occupant
void ElevatorControlSystem::handlePassengerBoardingElevator(Passenger* passenger, int elevator)
{
    if (passenger) {
        elevators[elevator]->addOccupant(passenger);
        floors[passenger->getCurrentFloor()]->removePerson(passenger);
        EventLogger::getInstance().logEvent(QString("Passenger %1 BOARDED elevator %2").arg(passenger->getId() + 1).arg(elevator + 1));
    }
}

// Move passenger over to floor
void ElevatorControlSystem::handlePassengerExitingElevator(Passenger* passenger, int elevator)
{
    if (passenger) {
        floors[passenger->getCurrentFloor()]->addPerson(passenger);
        elevators[elevator]->removeOccupant(passenger);
        EventLogger::getInstance().logEvent(QString("Passenger %1 EXITED elevator %2").arg(passenger->getId() + 1).arg(elevator + 1));
    }
}
