#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "elevatorcontrolsystem.h"
#include "eventlogger.h"

#include <QTabBar>
#include <QPushButton>
#include <QLCDNumber>
#include <QLabel>
#include <QDebug>

// SETTINGS
// I WOULD NOT CHANGE THESE AS THE UI IS STATIC
const int NUM_ELEVATORS = 3;
const int NUM_FLOORS = 7;
const int NUM_PASSENGERS = 3;

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
    , eventLogger(EventLogger::getInstance())
{
    ui->setupUi(this);
    // Hide tabs of floor boxes (little trick i did to have the effect)
    ui->floorWidget->tabBar()->hide();

    // Init eventLogger singlton
    eventLogger.setTextEdit(ui->eventLog);

    setupElevatorControlSystem();
    setupConnections();
}

MainWindow::~MainWindow()
{
    delete controlSystem;
    delete ui;
}

// Set up ECS
void MainWindow::setupElevatorControlSystem()
{
    controlSystem = new ElevatorControlSystem(setupElevators(), setupFloors(), setupPassengers(), this);
}

// Get elevator UI elements following patterning of layout
QVector<Elevator*> MainWindow::setupElevators()
{
    QVector<Elevator*> elevators;
    const int elevatorCapacity = ui->maxCapSlider->value();

    qInfo("\nCREATING %d ELEVATORS", NUM_ELEVATORS);

    for (int i = 0; i < NUM_ELEVATORS; ++i) {
        QFrame *elevatorFrame = qobject_cast<QFrame*>(ui->elevatorLayout->itemAt(i)->widget());
        if (!elevatorFrame) continue;

        qInfo("\nCreating ELEVATOR %d ...", i + 1);

        QLayout *elevatorButtonsLayout = elevatorFrame->findChild<QLayout*>();
        if (!elevatorButtonsLayout) continue;
        qInfo("Found elevator layout %s", elevatorButtonsLayout->objectName().toStdString().c_str());

        QLCDNumber *capacityDisplay = elevatorFrame->findChild<QLCDNumber*>();
        if (!capacityDisplay) continue;
        qInfo("Found elevator capacity display %s", capacityDisplay->objectName().toStdString().c_str());

        QLabel *elevatorDisplay = elevatorFrame->findChild<QLabel*>();
        if (!elevatorDisplay) continue;
        qInfo("Found elevator main display %s", elevatorDisplay->objectName().toStdString().c_str());

        QPushButton *openButton = qobject_cast<QPushButton*>(elevatorButtonsLayout->itemAt(0)->widget());
        QPushButton *closeButton = qobject_cast<QPushButton*>(elevatorButtonsLayout->itemAt(1)->widget());
        QPushButton *helpButton = qobject_cast<QPushButton*>(elevatorButtonsLayout->itemAt(2)->widget());
        QPushButton *obsButton = qobject_cast<QPushButton*>(elevatorButtonsLayout->itemAt(3)->widget());
        if(openButton) qInfo("Found open button %s", openButton->objectName().toStdString().c_str());
        if(closeButton) qInfo("Found close button %s", closeButton->objectName().toStdString().c_str());
        if(helpButton) qInfo("Found help button %s", helpButton->objectName().toStdString().c_str());
        if(obsButton) qInfo("Found obs button %s", obsButton->objectName().toStdString().c_str());

        Elevator *elevator = new Elevator(i, elevatorCapacity, openButton, closeButton, helpButton, obsButton, capacityDisplay, elevatorDisplay, this);
        elevators.append(elevator);
        qInfo("Created ELEVATOR %d", i + 1);
    }

    qInfo("\nDONE WITH ELEVATORS");
    return elevators;
}

// Get floor UI elements following patterning of layout
QVector<Floor*> MainWindow::setupFloors()
{
    QVector<Floor*> floors;

    qInfo("\nCREATING %d FLOORS", NUM_FLOORS);

    for (int i = 0; i < NUM_FLOORS; ++i) {
        QVector<QPushButton*> upButtons;
        QVector<QPushButton*> downButtons;

        qInfo("\nCreating FLOOR %d ...", i + 1);

        for (QLayout *passengerLayout : getPassengerTabLayouts()) {
            if (!passengerLayout) continue;
            qInfo("Found passenger tab layout %s", passengerLayout->objectName().toStdString().c_str());

            QLayout *floorButtonLayout = passengerLayout->itemAt((NUM_FLOORS - 1) - i)->widget()->findChild<QLayout*>();
            if (!floorButtonLayout) continue;
            qInfo("Found floor button layout %s", floorButtonLayout->objectName().toStdString().c_str());

            QLayout *directionButtonLayout = floorButtonLayout->itemAt(1)->layout();
            if (!directionButtonLayout) continue;
            qInfo("Found direction buttons layout %s", directionButtonLayout->objectName().toStdString().c_str());

            QPushButton *downButton = nullptr;
            QPushButton *upButton = nullptr;
            if(i == 0) {
                upButton = qobject_cast<QPushButton*>(directionButtonLayout->itemAt(0)->widget());
            } else if (i == NUM_FLOORS - 1) {
                downButton = qobject_cast<QPushButton*>(directionButtonLayout->itemAt(0)->widget());
            } else {
                downButton = qobject_cast<QPushButton*>(directionButtonLayout->itemAt(0)->widget());
                upButton = qobject_cast<QPushButton*>(directionButtonLayout->itemAt(1)->widget());
            }

            if (upButton) {
                upButtons.append(upButton);
                qInfo("Found up button %s", upButton->objectName().toStdString().c_str());
            }
            if (downButton) {
                downButtons.append(downButton);
                qInfo("Found down button %s", downButton->objectName().toStdString().c_str());
            }
        }

        Floor *floor = new Floor(i, upButtons, downButtons, this);
        floors.append(floor);
        qInfo("Created FLOOR %d", i + 1);
    }

    qInfo("\nDONE WITH FLOORS");
    return floors;
}

// Get passenger UI elements following patterning of layout
QVector<Passenger*> MainWindow::setupPassengers()
{
    QVector<Passenger*> passengers;

    qInfo("\nCREATING %d PASSENGERS", NUM_PASSENGERS);

    for (int i = 0; i < NUM_PASSENGERS; ++i) {
        QWidget *passengerControlWidget = ui->floorWidget->widget(i);
        if (!passengerControlWidget) continue;

        qInfo("\nCreating PASSENGER %d ...", i + 1);

        QLabel *passengerDisplay = passengerControlWidget->findChild<QLabel*>("personStatus" + QString::number(i + 1));
        if (!passengerDisplay) continue;

        qInfo("Found status label %s", passengerDisplay->objectName().toStdString().c_str());

        QVector<QPushButton*> floorSelectorButtons;
        QVector<QPushButton*> floorDirectionButtons;
        QVector<QPushButton*> personIndicators;

        for (QPushButton *button : passengerControlWidget->findChildren<QPushButton*>()) {
            if (button->objectName().startsWith("floorSelector")) {
                floorSelectorButtons.append(button);
                qInfo("Found floor selector button %s", button->objectName().toStdString().c_str());
            } else if (button->objectName().startsWith("floorDownBtn") || button->objectName().startsWith("floorUpBtn")) {
                floorDirectionButtons.append(button);
                qInfo("Found floor direction button %s", button->objectName().toStdString().c_str());
            } else if (button->objectName().startsWith("personIndicator")) {
                personIndicators.append(button);
                qInfo("Found passenger indicator %s", button->objectName().toStdString().c_str());
            }
        }

        Passenger *passenger = new Passenger(i, passengerDisplay, floorSelectorButtons, floorDirectionButtons, personIndicators, this);
        passengers.append(passenger);

        qInfo("Created PASSENGER %d", i + 1);
    }

    return passengers;
}

QVector<QLayout*> MainWindow::getPassengerTabLayouts()
{
    QVector<QLayout*> layouts;
    for (int i = 0; i < ui->floorWidget->count(); ++i) {
        layouts.append(ui->floorWidget->widget(i)->findChild<QLayout*>());
    }
    return layouts;
}

void MainWindow::setupConnections()
{
    for (int i = 0; i < ui->personLayout->count(); ++i) {
        QPushButton* personButton = qobject_cast<QPushButton*>(ui->personLayout->itemAt(i)->widget());
        if (personButton) {
            connect(personButton, &QPushButton::clicked, this, &MainWindow::handlePassengerButton);
        }
    }
}

void MainWindow::handlePassengerButton()
{
    QPushButton* senderButton = qobject_cast<QPushButton*>(sender());
    if (senderButton) {
        int index = ui->personLayout->indexOf(senderButton);
        ui->floorWidget->setCurrentIndex(index);
        for (int i = 0; i < ui->personLayout->count(); ++i) {
            QPushButton* personButton = qobject_cast<QPushButton*>(ui->personLayout->itemAt(i)->widget());
            if (personButton) {
                personButton->setChecked(i == index);
                personButton->setDisabled(i == index);
            }
        }
    }
}

// FIRE ALARM
void MainWindow::on_fireButton_clicked()
{
    eventLogger.logEvent("Fire alarm triggered");
    controlSystem->handleAlarm(AlarmType::FIRE);
}

// POWER OUT
void MainWindow::on_powerOutButton_clicked()
{
    eventLogger.logEvent("Power outage occurred");
    controlSystem->handleAlarm(AlarmType::POWEROUTAGE);
}

// CAPACITY CHANGED
void MainWindow::on_maxCapSlider_valueChanged(int value)
{
    eventLogger.logEvent("Maximum capacity changed to " + QString::number(value));
    for(Elevator* elevator : controlSystem->getElevators()) {
        elevator->setCapacity(value);
    }
}

// RESET
void MainWindow::on_resetButton_clicked()
{
    eventLogger.logEvent("Simulation Reset");
    controlSystem->reset();
}

