#include "passenger.h"
#include <QRegularExpression>
#include "eventlogger.h"

Passenger::Passenger(int id, QLabel* passengerStateDisplay, QVector<QPushButton*> desiredFloorButtons, QVector<QPushButton*> directionFloorButtons, QVector<QPushButton*> passengerIndicators, QObject* parent)
    : QObject(parent), id(id), currentFloor(0), destinationFloor(0), boardedElevator(-1), passengerStateDisplay(passengerStateDisplay), desiredFloorButtons(desiredFloorButtons), directionFloorButtons(directionFloorButtons), passengerIndicators(passengerIndicators)
{
    // Set floor
    setCurrentFloor(currentFloor);
    // Set up first pass UI
    updateButtonsAndIndicators(currentFloor + 1);

    // Add listners to all floor buttons
    for (QPushButton* button : desiredFloorButtons) {
        connect(button, &QPushButton::clicked, this, [=]() { handleDesiredFloorButtonClick(button, desiredFloorButtons, currentFloor); });
    }

}

Passenger::~Passenger()
{
}

int Passenger::getId() const
{
    return id;
}

int Passenger::getCurrentFloor() const
{
    return currentFloor;
}

int Passenger::getDestinationFloor() const
{
    return destinationFloor;
}

int Passenger::getBoardedElevator() const
{
    return boardedElevator;
}

void Passenger::setCurrentFloor(int floor)
{
    currentFloor = floor;
    emit floorChanged(floor);
}

void Passenger::lock() const
{
    for(QPushButton* b : directionFloorButtons) {
        b->setDisabled(true);
    }
    for(QPushButton* b : desiredFloorButtons) {
        b->setDisabled(true);
    }
}

void Passenger::setDestinationFloor(int floor)
{
    EventLogger::getInstance().logEvent(QString("Set destination to floor %1 for Passenger %2").arg(floor + 1).arg(id + 1));
    destinationFloor = floor;
    emit destinationFloorChanged(floor);
}

void Passenger::boardElevator(int elevator)
{
    passengerStateDisplay->setText("Boarding elevator");
    boardedElevator = elevator;
    updateButtonsAndIndicators(currentFloor + 1);
    emit boardingElevator(elevator);
}

void Passenger::exitElevator(int elevator)
{
    passengerStateDisplay->setText("Exiting elevator");
    boardedElevator = -1;
    updateButtonsAndIndicators(currentFloor + 1);
    emit exitingElevator(elevator);
}

void Passenger::updateButtonsAndIndicators(int newFloor)
{
    // UI goes from 1 onward not at 0 so this is often called with a +1
    // Disable all desired floor buttons
    for (QPushButton* button : desiredFloorButtons) {
        int floor = extractFloorFromButtonName(button->objectName());
        if (floor == newFloor && boardedElevator < 0) {
            button->setEnabled(false);
        } else {
            button->setEnabled(true);
        }
    }

    // Enable direction floor buttons for the new floor
    for (QPushButton* button : directionFloorButtons) {
        int floor = extractFloorFromButtonName(button->objectName());
        if (floor == newFloor && boardedElevator < 0) {
            button->setEnabled(true);
        } else {
            button->setEnabled(false);
        }
    }

    // Uncheck all passenger indicators
    for (QPushButton* indicator : passengerIndicators) {
        indicator->setChecked(false);
        indicator->setEnabled(false);
    }

    // Check the passenger indicator for the new floor
    for (QPushButton* indicator : passengerIndicators) {
        int floor = extractFloorFromButtonName(indicator->objectName());
        if (floor == newFloor && boardedElevator < 0) {
            indicator->setChecked(true);
            indicator->setEnabled(true);
            break;
        }
    }

    // Update passenger state display
    if (boardedElevator < 0) {
        passengerStateDisplay->setText(QString("Floor %1").arg(newFloor));
    } else {
        passengerStateDisplay->setText(QString("Elevator %1").arg(boardedElevator + 1));
    }
}

// Helper function to extract floor number from button name
// Not an ideal implementation as it would be better to have the QVectors ordered
// But I'm not sure if QT replicates the same order when finding UI elements on every machine
// So I did it this way to be safe
int Passenger::extractFloorFromButtonName(const QString& buttonName)
{
    // Assuming button names start with "floorButton" or "passengerIndicator" followed by the floor number
    // Used https://regex101.com/ to figure this out I initially had an ugly loop
    QRegularExpression regex("^\\D*?(\\d+)");
    QRegularExpressionMatch match = regex.match(buttonName);
    if (match.hasMatch()) {
        return match.captured(1).toInt();
    }

    // Pattern does not match
    return -1;
}

void Passenger::handleDesiredFloorButtonClick(QPushButton* button, const QVector<QPushButton*>& desiredFloorButtons, int currentFloor) {
    if (button->isChecked()) {
        // Minus one becuse the UI elements start at 1
        setDestinationFloor(extractFloorFromButtonName(button->objectName()) - 1);
        // Unhighlight all the other buttons
        for (QPushButton* otherButton : desiredFloorButtons) {
            otherButton->setChecked(otherButton == button);
        }
    } else {
        // If nothings checked assume passenger wants to get off at current floor
        setDestinationFloor(currentFloor);
    }
}
